{ **************************************************************************** }
{                                                                              }
{ Smart Mobile Studio - Runtime Library                                        }
{                                                                              }
{ Copyright (c) Eric Grange, Creative IT. All rights reserved.                 }
{ Licensed to Optimale Systemer AS.                                            }
{                                                                              }
{ **************************************************************************** }

unit GLS.Base;

interface

uses w3c.WebGL, w3c.TypedArray, w3c.DOM, GLS.Vectors;

type

   TGLRenderingContext = class
      private
         FHandle : JWebGLRenderingContext;

      public
         property GL : JWebGLRenderingContext read FHandle write FHandle;

   end;

   TGLContextObject = class abstract
      private
         FRC : TGLRenderingContext;
         FGL : JWebGLRenderingContext;

      public
         constructor Create(aRC : TGLRenderingContext); virtual;

         property RC : TGLRenderingContext read FRC;
         property GL : JWebGLRenderingContext read FGL;
   end;

   TGLBuffer = class abstract (TGLContextObject)
      protected
         FHandle : JWebGLBuffer;
         FBufferType : Integer;

      public
         constructor Create(aRC : TGLRenderingContext); override;
         destructor Destroy; override;

         procedure Bind;

         property Handle : JWebGLBuffer read FHandle;
   end;

   TGLArrayBufferUsage = (abuStream = JWebGLRenderingContext.STREAM_DRAW,
                          abuStatic = JWebGLRenderingContext.STATIC_DRAW,
                          abuDynamic = JWebGLRenderingContext.DYNAMIC_DRAW);

   TGLArrayBuffer = class (TGLBuffer)
      public
         constructor Create(aRC : TGLRenderingContext); override;

         procedure SetData(data : array of Float; usage : TGLArrayBufferUsage);

         procedure VertexAttribPointer(attrib, itemSize : Integer; normalized : Boolean; stride, offset : Integer);
   end;

   TGLElementArrayBuffer = class (TGLBuffer)
      public
         constructor Create(aRC : TGLRenderingContext); override;

         procedure SetData(data : array of Integer; usage : TGLArrayBufferUsage);
   end;

   TGLShader = class abstract (TGLContextObject)
      protected
         FHandle : JWebGLShader;

         procedure SetSource(src : String);
         function GetSource : String;

      public
         destructor Destroy; override;

         function Compile : Boolean; overload;
         function Compile(src : String) : Boolean; overload;
         function InfoLog : String;

         property Handle : JWebGLShader read FHandle;
         property Source : String read GetSource write SetSource;
   end;

   TGLFragmentShader = class (TGLShader)
      public
         constructor Create(aRC : TGLRenderingContext); override;
   end;

   TGLVertexShader = class (TGLShader)
      public
         constructor Create(aRC : TGLRenderingContext); override;
   end;

   TGLShaderProgram = class (TGLContextObject)
      private
         FHandle : JWebGLProgram;
         FVertex : TGLVertexShader;
         FFragment : TGLFragmentShader;
         FAttribInfos : array of JWebGLActiveInfo;
         FUniformInfos : array of JWebGLActiveInfo;
         FUniformLocations : array of JWebGLUniformLocation;

      protected
         procedure SetVertex(shader : TGLVertexShader);
         procedure SetFragment(shader : TGLFragmentShader);

         function GetAttribInfo(idx : Integer) : JWebGLActiveInfo;
         function GetUniformInfo(idx : Integer) : JWebGLActiveInfo;
         procedure CacheInfos;
         procedure ClearCacheInfos;

      public
         constructor Create(aRC : TGLRenderingContext); override;
         destructor Destroy; override;

         function Link : Boolean; overload;
         function Link(vs : TGLVertexShader; fs : TGLFragmentShader) : Boolean; overload;
         function Validate : Boolean;
         function InfoLog : String;

         function AttribLocation(attribName : String) : Integer;
         function UniformLocation(attribName : String) : JWebGLUniformLocation;

         property AttribInfo[idx : Integer] : JWebGLActiveInfo read GetAttribInfo;
         function AttribCount : Integer;
         property UniformInfo[idx : Integer] : JWebGLActiveInfo read GetUniformInfo;
         function UniformCount : Integer;

         procedure Use;

         procedure SetUniform(const name : String; i : Integer); overload;
         procedure SetUniform(const name : String; const m : Matrix4); overload;

         property Handle : JWebGLProgram read FHandle;
         property Vertex : TGLVertexShader read FVertex write SetVertex;
         property Fragment : TGLFragmentShader read FFragment write SetFragment;

   end;

implementation

// ------------------
// ------------------ TGLContextObject ------------------
// ------------------

constructor TGLContextObject.Create(aRC : TGLRenderingContext);
begin
   FRC:=aRC;
   FGL:=aRC.GL;
end;

// ------------------
// ------------------ TGLBuffer ------------------
// ------------------

constructor TGLBuffer.Create(aRC : TGLRenderingContext);
begin
   inherited Create(aRC);
   FHandle:=GL.createBuffer;
end;

destructor TGLBuffer.Destroy;
begin
   GL.deleteBuffer(Handle);
   inherited Destroy;
end;

procedure TGLBuffer.Bind;
begin
   GL.bindBuffer(FBufferType, Handle);
end;

// ------------------
// ------------------ TGLArrayBuffer ------------------
// ------------------

constructor TGLArrayBuffer.Create(aRC : TGLRenderingContext);
begin
   inherited Create(aRC);
   FBufferType:=GL.ARRAY_BUFFER;
end;

procedure TGLArrayBuffer.SetData(data : array of Float; usage : TGLArrayBufferUsage);
begin
   Bind;
   GL.bufferData(FBufferType, new JFloat32Array(data), usage);
end;

procedure TGLArrayBuffer.VertexAttribPointer(attrib, itemSize : Integer; normalized : Boolean; stride, offset : Integer);
begin
   Bind;
   GL.vertexAttribPointer(attrib, itemSize, GL.FLOAT, normalized, stride, offset);
end;

// ------------------
// ------------------ TGLElementArrayBuffer ------------------
// ------------------

constructor TGLElementArrayBuffer.Create(aRC : TGLRenderingContext);
begin
   inherited Create(aRC);
   FBufferType:=GL.ELEMENT_ARRAY_BUFFER;
end;

procedure TGLElementArrayBuffer.SetData(data : array of Integer; usage : TGLArrayBufferUsage);
begin
   Bind;
   GL.bufferData(FBufferType, new JUInt16Array(data), usage);
end;

// ------------------
// ------------------ TGLShader ------------------
// ------------------

destructor TGLShader.Destroy;
begin
   GL.deleteShader(Handle);
   inherited Destroy;
end;

procedure TGLShader.SetSource(src : String);
begin
   GL.shaderSource(Handle, src);
end;

function TGLShader.GetSource : String;
begin
   Result:=GL.getShaderSource(Handle);
end;

function TGLShader.Compile : Boolean;
begin
   GL.compileShader(Handle);
   Result:=GL.getShaderParameter(Handle, GL.COMPILE_STATUS);
end;

function TGLShader.Compile(src : String) : Boolean;
begin
   Source:=src;
   Result:=Compile;
end;

function TGLShader.InfoLog : String;
begin
   Result:=GL.getShaderInfoLog(Handle);
end;

// ------------------
// ------------------ TGLFragmentShader ------------------
// ------------------

constructor TGLFragmentShader.Create(aRC : TGLRenderingContext);
begin
   inherited Create(aRC);
   FHandle:=GL.createShader(GL.FRAGMENT_SHADER);
end;

// ------------------
// ------------------ TGLVertexShader ------------------
// ------------------

constructor TGLVertexShader.Create(aRC : TGLRenderingContext);
begin
   inherited Create(aRC);
   FHandle:=GL.createShader(GL.VERTEX_SHADER);
end;

// ------------------
// ------------------ TGLShaderProgram ------------------
// ------------------

constructor TGLShaderProgram.Create(aRC : TGLRenderingContext);
begin
   inherited Create(aRC);
   FHandle:=GL.createProgram;
end;

destructor TGLShaderProgram.Destroy;
begin
   Vertex:=nil;
   Fragment:=nil;
   GL.deleteProgram(Handle);
   inherited Destroy;
end;

procedure TGLShaderProgram.SetVertex(shader : TGLVertexShader);
begin
   if Vertex<>nil then
      GL.detachShader(Handle, Vertex.Handle);
   FVertex:=shader;
   if shader<>nil then
      GL.attachShader(Handle, shader.Handle);
end;

procedure TGLShaderProgram.SetFragment(shader : TGLFragmentShader);
begin
   if Fragment<>nil then
      GL.detachShader(Handle, Fragment.Handle);
   FFragment:=shader;
   if shader<>nil then
      GL.attachShader(Handle, shader.Handle);
end;

function TGLShaderProgram.Link : Boolean;
begin
   GL.linkProgram(Handle);
   Result:=gl.getProgramParameter(Handle, gl.LINK_STATUS);
   if Result then
      CacheInfos;
end;

function TGLShaderProgram.Link(vs : TGLVertexShader; fs : TGLFragmentShader) : Boolean;
begin
   Vertex:=vs;
   Fragment:=fs;
   Result:=Link;
end;

function TGLShaderProgram.Validate : Boolean;
begin
   GL.validateProgram(Handle);
   Result:=gl.getProgramParameter(Handle, gl.VALIDATE_STATUS);
end;

function TGLShaderProgram.InfoLog : String;
begin
   Result:=GL.getProgramInfoLog(Handle);
end;

procedure TGLShaderProgram.CacheInfos;
var
   i : Integer;
begin
   ClearCacheInfos;
   for i:=0 to Integer(GL.getProgramParameter(Handle, GL.ACTIVE_ATTRIBUTES))-1 do
      FAttribInfos.Add(GL.getActiveAttrib(Handle, i));
   for i:=0 to Integer(GL.getProgramParameter(Handle, GL.ACTIVE_UNIFORMS))-1 do begin
      var info := GL.getActiveUniform(Handle, i);
      FUniformInfos.Add(info);
      FUniformLocations.Add(GL.getUniformLocation(Handle, info.name));
   end;
end;

procedure TGLShaderProgram.ClearCacheInfos;
begin
   FAttribInfos.Clear;
   FUniformInfos.Clear;
   FUniformLocations.Clear;
end;

function TGLShaderProgram.AttribLocation(attribName : String) : Integer;
begin
   for Result:=FAttribInfos.High downto 0 do
      if FAttribInfos[Result].name=attribName then
         Exit;
   Result:=-1;
end;

function TGLShaderProgram.UniformLocation(attribName : String) : JWebGLUniformLocation;
var
   i : Integer;
begin
   for i:=FUniformInfos.High downto 0 do
      if FUniformInfos[i].name=attribName then
         Exit(FUniformLocations[i]);
end;

function TGLShaderProgram.GetAttribInfo(idx : Integer) : JWebGLActiveInfo;
begin
   Result:=FAttribInfos[idx];
end;

function TGLShaderProgram.AttribCount : Integer;
begin
   Result:=FAttribInfos.Length;
end;

function TGLShaderProgram.GetUniformInfo(idx : Integer) : JWebGLActiveInfo;
begin
   Result:=FUniformInfos[idx];
end;

function TGLShaderProgram.UniformCount : Integer;
begin
   Result:=FUniformInfos.Length;
end;

procedure TGLShaderProgram.Use;
begin
   GL.useProgram(Handle);
end;

procedure TGLShaderProgram.SetUniform(const name : String; i : Integer);
begin
   GL.uniform1i(UniformLocation(name), i);
end;

procedure TGLShaderProgram.SetUniform(const name : String; const m : Matrix4);
begin
   GL.uniformMatrix4fv(UniformLocation(name), False, m.ToFloat32Array);
end;

